import logging
import os
import time
from datetime import datetime
from multiprocessing import Pool
from pathlib import Path
from random import random
from typing import List, Union, Optional, Tuple

import grid2op
import numpy as np
from grid2op.Agent import BaseAgent
from grid2op.Environment import BaseEnv
from matplotlib import pyplot as plt

from curriculumagent.common.utilities import split_and_execute_action, find_best_line_to_reconnect
from curriculumagent.tutor.tutors.general_tutor import GeneralTutor


def collect_topology_tutor_one_chronic(
        action_paths: Union[Path, List[Path]],
        chronics_id: int,
        env_name_path: Union[Path, str] = "l2rpn_neurips_2020_track1_small",
        seed: Optional[int] = None,
        enable_logging: bool = True,
        TutorAgent: BaseAgent = GeneralTutor,
        tutor_kwargs: Optional[dict] = {},
        env_kwargs: Optional[dict] = {}
) -> Tuple[np.ndarray, np.ndarray]:
    """Collect tutor experience of one chronic.

    Args:
        action_paths: List of Paths for the tutor. This can be set to None if you want to use the
        DoNothingAgent for testing.
        chronics_id: Number of chronic to run.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        seed: Whether to init the Grid2Op env with a seed
        enable_logging: Whether to log the Tutor experience search.
        subset: Optional argument, whether the observations should be filtered when saved.
            The default version saves the observations according to obs.to_vect(), however if
            subset is set to True, then only the all observations regarding the lines, busses, generators and loads are
            selected. Further note, that it is possible to say "graph" in order to get the connectivity_matrix as well.
        TutorAgent: Tutor Agent which should be used for the search.
        tutor_kwargs: Additional arguments for the tutor, e.g. the max rho or the topology argument.
        env_kwargs: Optional arguments that should be used when initializing the environment.

    Returns:
        None.
    """
    if enable_logging:
        logging.basicConfig(level=logging.INFO)

    try:
        # if lightsim2grid is available, use it.
        from lightsim2grid import LightSimBackend

        backend = LightSimBackend()
        env = grid2op.make(dataset=env_name_path, backend=backend, **env_kwargs)
    except ImportError:  # noqa
        env = grid2op.make(dataset=env_name_path, **env_kwargs)
        logging.warning("Not using lightsim2grid! Operation will be slow!")

    if seed:
        env.seed(seed)

    env.set_id(chronics_id)
    env.reset()
    logging.info(f"current chronic:{env.chronics_handler.get_name()}")

    # After initializing the environment, let's init the tutor
    if tutor_kwargs is None:
        tutor_kwargs = {}
    else:
        logging.info(f"Run Tutor with these additional kwargs {tutor_kwargs}")

    if action_paths is not None:
        tutor_kwargs["action_space_file"] = action_paths

    tutor = TutorAgent(action_space=env.action_space, **tutor_kwargs)
    # first col for label which is the action index, remaining cols for feature (observation.to_vect())
    done, step, obs = False, 0, env.get_obs()

    vect_obs = obs.to_vect()
    # action ID + vector of observation
    obs_records = np.empty((0, 1 + len(vect_obs) + len(env.action_space().to_vect())), dtype=np.float32)

    # Count, topo_vect
    # Start with the default topology
    topo_records = np.hstack((0, obs.topo_vect)).reshape(1,-1)

    vect_obs = None
    a_comb = env.action_space({})
    while not done:
        # Get Tutor action
        action = tutor.act(observation=obs, reward=0, done=done)

        if isinstance(action, np.ndarray):
            action: grid2op.Action.BaseAction = tutor.action_space.from_vect(action)

        # Option1: Tutor is active, let's save the observation and execute the action
        if action.as_dict() != {}:

            # 3. Save vect_obs and add actions. Save only if first action of tutor in this run.
            if a_comb.as_dict() == {}:
                vect_obs = obs.to_vect().copy()
            a_comb += action

            # Execute Action:
            # This method does up to three steps and returns the output
            obs, _, done, _ = split_and_execute_action(env=env, action_vect=action.to_vect())

        # Option2: The Tutor produces a do nothing action. This happends, because the grid is stable
        # or it doesn't find an adequate method. One way or the other, we save the topology vector and
        # the observation!
        else:
            # First, we have to save the topo records and count the number of "do-nothing" actions of the tutor:
            topo_id = np.where((obs.topo_vect == topo_records[:, 1:]).all(axis=1))[0]

            if len(topo_id) > 0:
                topo_records[topo_id, 0] = topo_records[topo_id, 0] + 1
            else:
                topo_records = np.concatenate(
                    (topo_records, np.hstack([1, obs.topo_vect]).reshape(1, -1)), axis=0)
                # Set Topo_id:
                topo_id = len(topo_records)

            # Second, if this do nothing happened due to a tutor action, we expect it to save the
            # topology action of the topo_records with the observation and action !
            if a_comb.as_dict() != {}:
                # Save obs_record & set a_comb back to 0
                obs_records = np.concatenate(
                    (obs_records, np.hstack([topo_id, vect_obs, a_comb.to_vect()]).astype(np.float32).reshape(1, -1)),
                    axis=0)

                a_comb = env.action_space({})

            # Use Do-Nothing Action
            act_with_line = find_best_line_to_reconnect(obs, env.action_space({}))
            obs, _, done, _ = env.step(act_with_line)

        step = env.nb_time_step

    logging.info(f"game over at step-{step} with a total of {len(topo_records)} different topologies.")

    return topo_records, obs_records


def generate_topo_tutor_experience(
        env_name_path: Union[Path, str],
        save_path: Union[Path, str],
        action_paths: Union[Path, List[Path]],
        num_chronics: Optional[int] = None,
        num_sample: Optional[int] = None,
        jobs: int = -1,
        seed: Optional[int] = None,
        TutorAgent: BaseAgent = GeneralTutor,
        tutor_kwargs: Optional[dict] = {},
        env_kwargs: Optional[dict] = {},
):
    """Method to run the Tutor in parallel.

    Args:
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        save_path: Where to save the experience.
        action_paths: List of action sets (in .npy format).
        num_chronics: Total numer of chronics.
        num_sample: Length of sample from the num_chronics. If num_sample is smaller than num chronics,
            a subset is taken. If it is larger, the chronics are sampled with replacement.
        subset: Optional argument, whether the observations should be filtered when saved.
            The default version saves the observations according to obs.to_vect(), however if
            subset is set to True, then only the all observations regarding the lines, busses, generators and loads are
            selected. Further, note that it is possible to say graph in order to get the connectivity_matrix as well.
        jobs: Number of jobs in parallel.
        seed: Whether to set a seed to the sampling of environments
        TutorAgent: Tutor Agent which should be used for the search, default is the GeneralTutor.
        tutor_kwargs: Optional arguments that should be passed to the tutor agent.
        env_kwargs: Optional arguments that should be used when initializing the environment.

    Returns:
        None, saves results as numpy file.

    """
    log_format = "(%(asctime)s) [%(name)-10s] %(levelname)8s: %(message)s [%(filename)s:%(lineno)s]"
    logging.basicConfig(level=logging.INFO, format=log_format)

    if jobs == -1:
        jobs = os.cpu_count()

    tasks = []

    # Make sure we can initialize the environment
    # This also makes sure that the environment actually exits or gets downloaded
    env: BaseEnv = grid2op.make(env_name_path)
    chronics_path = env.chronics_handler.path
    if chronics_path is None:
        raise ValueError(f"Can't determine chronics path of given environment {env_name_path}")

    if num_chronics is None:
        num_chronics = len(os.listdir(chronics_path))

    if num_sample:
        if num_sample <= num_chronics:
            sampled_chronics = random.sample(range(num_chronics), num_sample)
        else:
            sampled_chronics = random.choices(np.arange(num_chronics), k=num_sample)
    else:
        sampled_chronics = np.arange(num_chronics)

    for chronic_id in sampled_chronics:
        tasks.append((action_paths, chronic_id, env_name_path, seed, True, TutorAgent, tutor_kwargs,env_kwargs))
    if jobs == 1:
        # This makes debugging easier since we don't fork into multiple processes
        logging.info(f"The following {len(tasks)} tasks will executed sequentially: {tasks}")
        out_result = []
        for task in tasks:
            out_result.append(collect_topology_tutor_one_chronic(*task))
    else:
        logging.info(f"The following {len(tasks)} tasks will be distributed to a pool of {jobs} workers:")
        start = time.time()
        with Pool(jobs) as p:
            # This now countains touples:
            out_result = p.starmap(collect_topology_tutor_one_chronic, tasks)
        end = time.time()
        elapsed = end - start
        logging.info(f"Time: {elapsed}s")

    ############# Combine the results BUT also adjust the respective Ids!
    # Now concatenate the result:

    logging.info("Running through the individual runs per scenario and combining them to one"
                 "file\n")
    # Save beginning files
    global_topo_id, global_obs_records = combine_topo_datasets(out_result)

    logging.info(f"Lenght of topologies is {len(global_topo_id)}")
    logging.info(f"Lenght of obsact is {len(global_obs_records)}")

    if save_path.is_dir():
        now = datetime.now().strftime("%d%m%Y_%H%M%S")
        save_path = save_path / f"tutor_experience_{now}.npz"

    np.savez_compressed(save_path, topo_id=global_topo_id, obs_records=global_obs_records)
    logging.info(f"Tutor experience has been saved to {save_path}")


def combine_topo_datasets(comb_res:list)->Tuple[np.ndarray,np.ndarray]:
    """ Runs through all data of the provided list and then combines both the topology action as well
    as the global_obs_records

    Args:
        comb_res: Output of previously runs in a list. Within the list, each run should hav the order (topo_id,
        obs_records).

    Returns:

    """
    global_topo_id = comb_res[0][0].copy()
    global_obs_records = comb_res[0][1].copy()

    for dat in comb_res[1:]:
        topo_records, obs_records = dat

        # First check, whether the topologies are already in the global_topo_id
        key_tuples = []
        for i, topo in enumerate(topo_records):
            topo_id = np.where((topo[1:] == global_topo_id[:, 1:]).all(axis=1))[0]
            # The topology already exists:
            if len(topo_id) > 0:
                key_tuples.append((int(i), int(topo_id)))
                global_topo_id[topo_id, 0] += topo[0]
            else:
                global_topo_id = np.vstack((global_topo_id, topo))

        # Now we overwrite the obsact record an save it:
        for i, j in key_tuples:
            obs_records[obs_records[:, 0] == i, 0] = j

        global_obs_records = np.vstack((global_obs_records, obs_records))
    return global_topo_id,global_obs_records

def prepare_topo_dataset(
        traindata_path: Path,
        target_path: Path,
        dataset_name: str,
        seed: Optional[int] = 42,
        top_k_topologies:int = None,
        plot_choice:bool=False,
        env:grid2op.Environment =None
):
    """ A specific dataset to load all required data / action types and save them again into training/val/testing

    Args:
        traindata_path: The path to the generated records_* files by the tutor to use.
        target_path: The path to save the dataset files to.
        dataset_name: Name of the dataset which gets used to create the files.
        seed: The random seed for shuffling the samples.
        top_k_topologies: whether to only return a given set of topologies,i.e., only look at the top k of topologies.
        If set to none, we look at all topologies.
        plot_choice: Whether to plot the number of topolgies
        env: provide a grid2op Env to only extract the observation (and not the action as well)

    Returns:
        None.

    """
    np.random.seed(seed)

    assert traindata_path.exists() and traindata_path.is_dir(), f"{traindata_path} is not a valid directory"

    # Change this more globaly
    npy_files = list(traindata_path.glob("*.npz"))
    assert len(npy_files), f"There are no files at {traindata_path}"

    logging.info(f"Creating final dataset at directory {target_path}")

    npy_files = sorted(npy_files)  # This will hopefully make this reproducible

    loaded_actions = [np.load(str(f)) for f in npy_files]
    extracted_dat = [[f["topo_id"],f["obs_records"]] for f in loaded_actions]
    global_topo_id, global_obs_records = combine_topo_datasets(extracted_dat)

    logging.info(f"Lenght of topologies is {len(global_topo_id)}")
    logging.info(f"Lenght of obsact is {len(global_obs_records)}")

    # Topologies:
    top_freq = global_topo_id[global_topo_id[:, 0].argsort()[::-1],0]

    if plot_choice:
        plt.plot(top_freq[:1000])
        if top_k_topologies:
            plt.vlines(top_k_topologies, ymin=0, ymax=top_freq.max(), color="red")
        plt.title("Sorted topology based on of the top 1000 topologies")
        plt.legend(["Sorted frequency of topology", "best_n"])
        plt.show()

    if top_k_topologies:
        # Now extract the subset:
        logging.info(f"Only considering the top k topologies with {top_k_topologies}")
        top_k = global_topo_id[:, 0].argsort()[-top_k_topologies:]
        mask = [i in top_k for i in global_obs_records[:, 0]]
        sub_topo_ids = global_topo_id[top_k[::-1]]
        sub_obs_records = global_obs_records[mask]
        for i, j in enumerate(top_k[::-1]):
            mask = sub_obs_records[:, 0] == j
            sub_obs_records[mask, 0] = i
        unique_obs = np.unique(sub_obs_records,axis=0)

        logging.info(f"We have a total of {sub_obs_records.shape[0]} observation, where {unique_obs.shape[0]} "
                     f"are unique. This is a total of {unique_obs.shape[0]/sub_obs_records.shape[0]}%")

    else:
        unique_obs = np.unique(global_obs_records, axis=0)

        logging.info(f"We have a total of {global_obs_records.shape[0]} observation, where {unique_obs.shape[0]} "
                     f"are unique. This is a total of {unique_obs.shape[0] / global_obs_records.shape[0]}%")
        sub_topo_ids = global_topo_id


    # Now let's save the observation and action sets:
    np.random.shuffle(unique_obs)

    assert target_path.exists() and target_path.is_dir(), "Given target_path is not a directory that exists"

    # Dataset partition
    num_sampling = unique_obs.shape[0]
    train_size = num_sampling * 8 // 10
    # validate_size = num_sampling // 10
    test_size = num_sampling // 10
    # s for state, feature; a for action, label.

    if env:
        max_steps = env.get_obs().to_vect().shape[0] +1
    else:
        max_steps = None

    # Check whether the input is a Tuple ID or not. If yes
    s_train, a_train = unique_obs[:train_size, 1:max_steps], unique_obs[:train_size, :1].astype(int)
    s_validate, a_validate = unique_obs[train_size:-test_size, 1:max_steps], unique_obs[train_size:-test_size, :1].astype(int)
    s_test, a_test = unique_obs[-test_size:, 1:max_steps], unique_obs[-test_size:, :1].astype(int)

    logging.info(f"TrainSet Size: {len(s_train)}")
    logging.info(f"ValidationSet Size: {len(s_validate)}")
    logging.info(f"TestSet Size: {len(s_test)}")

    train_path = target_path / f"{dataset_name}_train.npz"
    np.savez_compressed(train_path, s_train=s_train, a_train=a_train)

    validation_path = target_path / f"{dataset_name}_val.npz"
    np.savez_compressed(validation_path, s_validate=s_validate, a_validate=a_validate)

    test_path = target_path / f"{dataset_name}_test.npz"
    np.savez_compressed(test_path, s_test=s_test, a_test=a_test)

    np.save(target_path / "topologies.npy",sub_topo_ids)


"""This file contains the N-1 Teacher which tries to find actions that make the power network more robust to attacks.

For this teacher it is possible to enforce some line disconnection.
"""
import logging
import multiprocessing
import os
import time
from copy import deepcopy
from multiprocessing import Pool
from pathlib import Path
from pprint import pformat
from typing import Union, Optional, List, Tuple

import defopt
import grid2op
import numpy as np
from grid2op.Action import BaseAction
from grid2op.Backend import PandaPowerBackend
from grid2op.Environment import BaseEnv
from grid2op.Observation import BaseObservation

from curriculumagent.teacher.submodule.common import save_sample_new
from curriculumagent.common.utilities import find_best_line_to_reconnect
from curriculumagent.teacher.submodule.encoded_action import EncodedTopologyAction
from curriculumagent.teacher.submodule.topology_action_search import topology_search_topk


class NMinusOneTeacher:
    """Class to collect the N-1 data.

    """

    def __init__(
            self,
            lines_to_attack: Optional[List[float]] = None,
            rho_n0_threshold: Optional[float] = 0.9,
            rho_max_threshold: Optional[float] = 1.0,
            seed: Optional[int] = None,
    ):
        """Constructor method of the N-1 Teacher.

        Difference between rho_n0_threshold and rho_max_threshold is that for the first,
        the N-1 Search starts, while the latter is necessary to just use the best greedy action.

        Args:
            lines_to_attack: Lines to select for the N-1 check.
            rho_n0_threshold: Threshold, when the do-nothing agent should stop and start searching for N-1.
            rho_max_threshold: Threshold, when the N-1 search should stop and only the best greedy.
            action should be selected.
            seed: Optional seed for the Grid2Op Environment and random sampling.

        Returns:
            None.
        """
        if lines_to_attack is None:
            lines_to_attack = [45, 56, 0, 9, 13, 14, 18, 23, 27, 39]
        self.lines_to_attack = lines_to_attack
        self.rho_n0 = rho_n0_threshold
        self.rho_threshold = rho_max_threshold

        # Set internally:
        self.seed = seed
        self.top_k: int = 100
        self.all_actions: Optional[List[BaseAction]] = None  # Used to cache all_actions, see method below

    def get_all_actions(self, env: BaseEnv) -> List[BaseAction]:
        """Set and get all actions for the action space of the provided environment.

        Args:
            env: Grid2Op Environment

        Returns:
            All possible actions.

        """
        if self.all_actions is None:
            self.all_actions = env.action_space.get_all_unitary_topologies_set(env.action_space)

        return self.all_actions

    def do_nothing_and_run_through_lines_action(self, env: BaseEnv, obs: BaseObservation) -> BaseAction:
        """Method to check, whether a disconnection of the lines would result in to a breach of the
        rho_threshold.

        For this step, run through all lines and check which disconnection would result the most
        problematic rho value. If found, return that action.

        Note:
            Only values between the rho_n0 and the rho_max value are accepted.

        Args:
            env: Grid2Op environment.
            obs: Observation of the Grid2Op Environment.

        Returns:
            Action with problematic rho value.
        """
        cur_rho = self.rho_n0
        chosen_action = env.action_space({})
        # What does it do exactly?
        for line in self.lines_to_attack:
            new_line_status_array = np.zeros(obs.rho.shape, dtype=int)
            new_line_status_array[line] = -1
            action = env.action_space({"set_line_status": new_line_status_array})
            obs_sim, _, done, _ = obs.simulate(action)
            if (obs_sim.rho.max() > cur_rho) and obs_sim.rho.max() < self.rho_threshold and done is False:
                cur_rho = obs_sim.rho.max()
                chosen_action = action
        return chosen_action

    def calculate_attacked_max_rho(self, obs: BaseObservation, action: BaseAction) -> float:
        """Method to run through all lines and check the maximum rho  if the lines is disconnected.

        This is to measure how robust the supplied action is against the attacks, the lower, the better.

        Note:
            If the game is done, the rho will be set to np.inf.

        Args:
            obs: Observation of the Grid2Op Environment.
            action: Current selected action.

        Returns:
            The maximum rho of the action.

        """
        p_max_max = 0
        for line in self.lines_to_attack:
            new_line_status_array = np.zeros(obs.rho.shape, dtype=int)
            new_line_status_array[line] = -1
            disconnect_action = deepcopy(action)
            disconnect_action.update({"set_line_status": new_line_status_array})
            obs_new, _, done, _ = obs.simulate(disconnect_action)
            rho_max = obs_new.rho.max()
            if rho_max > p_max_max and done is False:
                p_max_max = rho_max
            if done:
                p_max_max = np.inf

        return p_max_max

    def search_best_n_minus_one_action(self, env: BaseEnv, obs: BaseObservation) -> Tuple[BaseAction, bool]:
        """Method to collect the best actions that might be N-1.

        Args:
            env: The environment to collect the actions in.
            obs: Observation of the Grid2Op Env.

        Returns:
            The selected action as well as boolean which indicates whether the action was N-1.
            The selected action can either be the N-1 action or the best greedy action.
        """
        # First select the best greedy actions:
        all_actions = self.get_all_actions(env)
        greedy_action_set = topology_search_topk(env, obs, all_actions, top_k=self.top_k)
        # No good action found -> abort
        if len(greedy_action_set) == 0:
            return env.action_space({}), False

        # Now run simulate then N-1 actions

        # Only consider actions that decrease the max_rho below rho_n0
        action_set = [new_act for rho_delta, new_act in greedy_action_set if (obs.rho.max() - rho_delta) < self.rho_n0]

        n_1_action_found = False
        if action_set:
            chosen_action = action_set[0]
            rho_max_max = np.inf  # Note: This can be problematic?
            for act in action_set:
                current_rho_max_max = self.calculate_attacked_max_rho(obs=obs, action=act)
                if current_rho_max_max < rho_max_max:
                    rho_max_max = current_rho_max_max
                    chosen_action = act
                    n_1_action_found = True

        if not n_1_action_found:
            chosen_action = greedy_action_set[0][1]

        return chosen_action, n_1_action_found

    def n_minus_one_agent(
            self,
            env_path: str,
            chronics_path: str,
            chronics_id: Union[str, int],
            save_path: Path,
            save_greedy: bool = False,
            active_search: bool = False,
            disable_opponent: bool = False,
    ):

        """Running the N-1 Agent Search for specific chronic.

        For this purpose, we first reinitialize the environment, set a special chronics_id and
        then run the N-1 agent.

        Args:
            env_path: Path of the Grid2Op Environment as a str.
            chronics_path: Path of the chronics.
            chronics_id: ID of the chronic, which the env should be run on.
            save_path: Where to save the selected actions.
            save_greedy: Whether the greedy action should be saved, if False - only the N-1 actions are saved.
            active_search: Variable, whether the do-nothing agent should disconnect the lines.
            disable_opponent: Whether to disable the opponent of the environment or not. This is principally a
            good idea when you activate active search. In that case, only the N-1 agent should deactivate the
            lines.

        Returns:
            Action set that are N-1.
        """
        try:
            # if lightsim2grid is available, use it.
            from lightsim2grid import LightSimBackend

            backend = LightSimBackend()
        except ImportError:  # noqa
            backend = PandaPowerBackend()
            logging.warning("Not using lightsim2grid! Operation will be slow!")

        kwargs = dict()
        if disable_opponent:
            kwargs["opponent_budget_per_ts"] = 0
            kwargs["opponent_init_budget"] = 0

        if chronics_path:
            env = grid2op.make(dataset=env_path, chronics_path=chronics_path, backend=backend, **kwargs)
        else:
            env = grid2op.make(dataset=env_path, backend=backend, **kwargs)

        if self.seed:
            env.seed(self.seed)

        # Set chronic:
        env.set_id(chronics_id)
        env.reset()
        logging.info(f"[{env.chronics_handler.get_name()}] N-1 Search started")

        done = False
        obs = env.get_obs()
        # Run through episode:

        import warnings

        warnings.filterwarnings("ignore")

        while not done:
            save_action = False
            rho_max = obs.rho.max()
            if rho_max < self.rho_n0:  # < 0.9
                # Do Nothing and disconnect lines if necessary
                # Added only one line disconnect
                if active_search and all(obs.line_status):
                    # Why is it helping considering the most problematic action?
                    action = self.do_nothing_and_run_through_lines_action(env=env, obs=obs)
                else:
                    action = env.action_space({})
                    action = find_best_line_to_reconnect(obs, action)

            # If no line is disconnected and the rho lies below the threshold, we continue with the n-1 search.
            elif (rho_max < self.rho_threshold) and all(obs.line_status):  # < 1.0
                # Sort through N-1 actions becuse rho_n0 was breached:
                logging.info(
                    f"[{env.chronics_handler.get_name()}] ({obs.current_step}/{obs.max_step}) (rho={rho_max}): Looking for N-1 action"
                )
                action, n_1_action_found = self.search_best_n_minus_one_action(env=env, obs=obs)
                action = find_best_line_to_reconnect(obs, action)
                if n_1_action_found or (n_1_action_found is False and save_greedy):
                    save_action = True

            else:
                logging.info(
                    f"[{env.chronics_handler.get_name()}] ({obs.current_step}/{obs.max_step}) (rho={rho_max}): Looking for "
                    f"greedy action"
                )
                all_actions = self.get_all_actions(env)
                greedy_action_set = topology_search_topk(env, obs, all_actions, top_k=self.top_k)
                if len(greedy_action_set) == 0:
                    action = env.action_space({})
                    save_action = False
                else:
                    action = greedy_action_set[0][1]
                    save_action = save_greedy
                action = find_best_line_to_reconnect(obs, action)

            obs_, _, done, _ = env.step(action)
            # logging.info(f'[{env.chronics_handler.get_name()}] {obs.current_step}/{obs.max_step}: {done}')

            # Calculate rho_improvement for saving
            rho_max_old = obs.rho.max()
            rho_max_new = obs_.rho.max()
            rho_improvement: float = rho_max_old - rho_max_new

            if save_action:
                save_sample_new(Path(save_path), [(rho_improvement, EncodedTopologyAction(action))], obs, obs_, top_k=1)

            obs = obs_
        logging.info(f"[{env.chronics_handler.get_name()}] Search stopped at {obs.current_step}/{obs.max_step}")

    def collect_n_minus_1_experience(
            self,
            save_path: Path,
            env_name_path: Path,
            number_of_years: Optional[int] = None,
            jobs: int = -1,
            save_greedy: bool = False,
            active_search: bool = True,
            disable_opponent: bool = True,
    ):

        """Collect useful actions by attacking/disconnecting specific lines at random timesteps
        and trying to find the best action to resolve that problem.

        Args:
            save_path: Path to save the teacher_experience file to.
            env_name_path: Path to Grid2Op dataset or the standard name of it.
            number_of_years: Optional number of years to restrict the cronics.
            jobs: Number of jobs to use for parallel teacher_experience collection. If set to -1, jobs is set
                to the number of CPUs
            save_greedy: Whether to save the greedy actions as well
            active_search: Should the N-1 agent be forced with line outages or not
            disable_opponent: Should the opponent be disabled or not

        Returns:
            None.
        """
        if jobs == -1:
            jobs = os.cpu_count()

        env_name_path = Path(env_name_path).expanduser()
        chronics_path = None
        if env_name_path.is_dir():
            chronics_path = str(env_name_path / "chronics")
        env_name_path = str(env_name_path)

        try:
            # if lightsim2grid is available, use it.
            from lightsim2grid import LightSimBackend

            backend = LightSimBackend()
        except ImportError:  # noqa
            backend = PandaPowerBackend()
            logging.warning("Not using lightsim2grid! Operation will be slow!")

        if chronics_path:
            # Use env_name_path as path together with chronics_path
            env = grid2op.make(dataset=env_name_path, chronics_path=chronics_path, backend=backend)
        else:
            # Use env_name_path as name
            env = grid2op.make(dataset=env_name_path, backend=backend)

        tasks = []
        if number_of_years:
            train_chronics = [
                chronic for chronic in env.chronics_handler.subpaths if int(chronic[-3:]) < number_of_years
            ]
        else:
            train_chronics = env.chronics_handler.subpaths

        for chronic_name in train_chronics:
            tasks.append(
                (env_name_path, chronics_path, chronic_name, save_path, save_greedy, active_search, disable_opponent)
            )
        logging.info(
            f"These {len(tasks)} scenarios will be trained using a pool of {jobs} workers:\n {pformat(train_chronics)}"
        )
        start = time.time()
        if jobs == 1:
            for task in tasks:
                self.n_minus_one_agent(*task)

        with Pool(jobs) as p:
            p.starmap(self.n_minus_one_agent, tasks)

        end = time.time()
        elapsed = end - start
        logging.info(f"Time: {elapsed}s")






"""Setup-script for curriculumagent package."""
import os
import sys
import importlib
from setuptools import setup, find_packages
import versioneer

PACKAGE_NAME = 'curriculumagent'  # name used by PIP
IMPORT_NAME = 'curriculumagent'  # name used for imports
NAMESPACE = ""

with open('requirements.txt', 'r') as f:
    requirements_install = f.readlines()

# Add all the packages in requirements.txt to an extra named '[dev]'
with open('requirements-dev.txt', 'r') as f:
    requirements_dev = f.readlines()

# All the requirements
requirements = {
    'install': [
        r for r in requirements_install if not r.startswith('-r ')
    ],
    'test': [  # testing
        'pytest',
        'coverage',
    ],
    'docs': [  # documentation generation and tools
        'Sphinx',
        'sphinx-rtd-theme',  # nicer theme for the html docs
        'sphinx-autodoc-typehints',  # use typehints to provide types
    ],
    'dev': [
        r for r in requirements_dev if not r.startswith('-r ')
    ],
}

URLS = {'GITHUB': 'https://github.com/FraunhoferIEE/curriculumagent/',
        'Readthedocs': 'https://curriculumagent.readthedocs.io/en/latest/'
}
SCRIPTS = []  # no scripts necessary
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 3.10',
]

# ======================================================================================
# No configurable options below this line

# Get current version etc. from the `__about__.py` file --------------------------------
# Very evil way to retrieve the variables contained in about.
# Those will be added to the current scope
PROJECT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.append(PROJECT_PATH)
about = importlib.import_module(IMPORT_NAME + '.__about__')

# def find_version(*file_paths):
#    """Function used to find the version of the package."""
#    version_file = read(*file_paths)
#    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
#                              version_file, re.MULTILINE)
#    if version_match:
#        return version_match.group(1)
#    raise RuntimeError("Unable to find version string for the package.")


if __name__ == '__main__':
    setup(
        # General ---
        name=PACKAGE_NAME,
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),

        packages=find_packages(exclude=['docs', 'tests', 'tests.*']),
        namespace_packages=([NAMESPACE] if NAMESPACE else []),
        scripts=SCRIPTS,

        install_requires=requirements['install'],
        extras_require={k: r for k, r in requirements.items() if k != 'install'},

        # Testing ---
        test_suite='pytest.collector',
        tests_require=requirements['test'],

        package_data={
            # If any package contains *.rst files, include them:
            '': ['*.rst', '*.npy'],
        },

        # Metadata --- (needed if we ever have a private Package Index)
        author=about.__author__,
        description=about.__description__,
        classifiers=CLASSIFIERS,
        long_description = open('README.md').read(),
        long_description_content_type = 'text/markdown',
        #long_description='CurriculumAgent is a cleanup and improved version of the NeurIPS 2020 Competition Agent by '
        #                 'Binbinchen.The agent is build to extract action sets of the Grid2Op Environment and then '
        #                 'use rule-based agent to train a Reinforcement Learning agent.',
        # license='???',
        # url='???'  # should link somewhere, perhaps to the docs
        project_urls=URLS,

        # could also include long_description, download_url, classifiers, etc.
    )



""" This file is the advanced agent constisting of a model that can cover different action spaces, rllib
training, scaling and other additional features.

To submit your own agent, just specify the model and action_space_path and if you want a scaler that
you previously trained with. The MyAgent will import the model and then act in the specific environment.

"""

import logging
import os
from pathlib import Path
from typing import Optional, Union, List

import grid2op
import numpy as np
import tensorflow as tf
import torch
from grid2op.Agent import BaseAgent
from sklearn.base import BaseEstimator
from tensorflow.keras.models import Model

from tests.data.training_data_track1.config import config
from ..junior.junior_student_pytorch import Junior_PtL

try:
    # Import if the agent is coppied into the submission folder
    from curriculumagent.common.obs_converter import obs_to_vect
    from curriculumagent.common.utilities import (
        find_best_line_to_reconnect,
        is_legal,
        split_action_and_return,
        simulate_action, revert_topo,
    )

except ImportError:
    # Import if the agent is copied into the submission folder
    from .obs_converter import obs_to_vect
    from .utilities import (
        find_best_line_to_reconnect,
        is_legal,
        split_action_and_return,
        simulate_action, revert_topo,
    )


class MyAgent(BaseAgent):
    def __init__(
            self,
            action_space,
            model_path: Union[Path, str],
            action_space_path: Optional[Union[Path, List[Path]]] = None,
            this_directory_path: Optional[str] = "./",
            subset: Optional[bool] = False,
            scaler: Optional[BaseEstimator] = None,
            best_action_threshold: float = 0.95,
            topo: Optional[bool] = False,
            check_overload: Optional[bool] = False,
            max_action_sim: Optional[int] = 50,
            action_space_file: Optional[str] = None,
            run_with_tf: bool = True,
    ):
        """The new advanced agent.

        In contrast to the original agent, this agent enables the implementation of tuple and triple actions,
        as well as use either a keras model.

        Next to the difference in the models and actions, this agent also has the ability to transform the
        observations based on a provided scaler and/or filter them accordingly.

        Note:
            If you just want to pass this agent as submission without the CurriculumAgent, copy the content
            of the common dir into this directory. Further, add the model and actions to complete it.

        Args:
            action_space: Action Space of the Grid2Op Enviornment
            model_path: Path where to find the rllib model or Keras model
            action_space_path: path, where to find the action sets. This is required to run the agent
            this_directory_path: Path of the submission directory
            subset: Boolean, whether to filter the observation
            scaler: Optional Scaler for the neural network
            best_action_threshold: Threshold, when to stop searching for the results.
            topo: Booling indicator, whether the agent should revert to original topology if it is possible
            check_overload: Boolean, whether to simulate a stress of the generation and load
            max_action_sim: Define, how many of the actions you want to evaluate before selecting a suitable
            candidate. If you want to select all, it has to be the number of actions. For a more rapid simulation, you
            can just select fewer values.
            action_space_file: Optional alternative to action_space_path, if you want to provide the file itself.
            run_with_tf: Flag to determine if the agent should use TensorFlow (True) or PyTorch (False).
        """
        # Initialize a new agent.
        BaseAgent.__init__(self, action_space=action_space)

        # Collect action set:
        # If action_space_file is available, we overwrite it here
        if isinstance(action_space_file,(str,Path)):
            action_space_path = Path(action_space_file)

        self.actions = self.__collect_action(
            this_directory_path=this_directory_path, action_space_path=action_space_path
        )

        self.subset = subset
        self.check_overload = check_overload
        self.run_with_tf = run_with_tf

        # Initialize model based on the framework
        if self.run_with_tf:
            # TensorFlow model loading
            try:
                self.model = tf.keras.models.load_model(model_path, compile=False)
                self.model.compile()
                logging.info(f"Successfully loaded TensorFlow model from {model_path}.")
            except (IndexError, AttributeError) as e:
                raise AttributeError(f"You are trying to load an old TensorFlow model that is not supported: {e}")
            except Exception as e:
                raise ValueError(f"Error loading TensorFlow model: {e}")
        else:
            # PyTorch model loading
            try:
                checkpoint = torch.load(model_path)
                if isinstance(checkpoint, dict):
                    # "Loading Junior model"
                    model = Junior_PtL.load_from_checkpoint(model_path)
                    self.model = model
                else:
                    # "Loading Senior model"
                    print("Loading Senior model")
                    self.model=checkpoint

                self.model.eval()  # Set model to evaluation mode
                logging.info(f"Successfully loaded PyTorch model from {model_path}.")
            except Exception as e:
                raise ValueError(f"Error loading PyTorch model: {e}")

        self.scaler = scaler
        self.recovery_stack = []
        self.overflow_steps = 0
        self.next_actions = None
        self.best_action_threshold = best_action_threshold
        self.max_action_sim = max_action_sim
        if topo:
            self.topo = topo
        else:
            self.topo = False

    def act_with_id(
            self, observation: grid2op.Observation.BaseObservation) -> grid2op.Action.BaseAction:
        """Method of the agent to act.

        When the function selects a tuple action or triple action, the next steps are predetermined as
        well,i.e., all actions are returned sequentially.

        Note: this method was primarily written to plug it into the generate experience methode

        Args:
            observation: Grid2Op Observation

        Returns: A suitable Grid2Op action

        """

        # Similar to the Tutor, we check whether there is some remaining action, based on previous
        # selected tuples
        if self.next_actions is not None:
            # Try to do a step:
            try:
                next_action = next(self.next_actions)
                next_action = find_best_line_to_reconnect(obs=observation, original_action=next_action)
                if is_legal(next_action, observation):
                    return next_action, -1
            except StopIteration:
                self.next_actions = None

        if observation.rho.max() >= 1:
            self.overflow_steps += 1
        else:
            self.overflow_steps = 0

        # case: secure with low threshold
        if observation.rho.max() < self.best_action_threshold:  # fixed threshold
            if self.topo:
                action_array = revert_topo(self.action_space, observation,
                                           rho_limit=0.8)
                default_action = self.action_space.from_vect(action_array)
                default_action = find_best_line_to_reconnect(obs=observation,
                                                             original_action=default_action)
            else:
                default_action = self.action_space({})
                default_action = find_best_line_to_reconnect(obs=observation,
                                                             original_action=default_action)

            return default_action, -1

        # Now, case dangerous:
        min_rho = observation.rho.max()

        logging.info(
            f"{observation.get_time_stamp()}s, heavy load,"
            f" line-{observation.rho.argmax()}d load is {observation.rho.max()}"
        )

        idx_chosen = None

        sorted_actions = self.__get_actions(obs=observation)[:self.max_action_sim]

        for k, idx in enumerate(sorted_actions):
            action_vect = self.actions[idx, :]
            rho_max, valid_action = simulate_action(action_space=self.action_space, obs=observation,
                                                    action_vect=action_vect, check_overload=self.check_overload
                                                    )
            if not valid_action:
                continue

            if rho_max <= self.best_action_threshold:
                # For a very suitable candidate, we break the loop
                logging.info(f"take action {idx}, max-rho to {rho_max}," f" simulation times: {k + 1}")
                idx_chosen = idx
                break

            if rho_max < min_rho:
                # If we have a decrease in rho, we already save the candidate.
                min_rho = rho_max
                idx_chosen = idx

        if idx_chosen:
            self.next_actions = split_action_and_return(observation, self.action_space, self.actions[idx_chosen, :])
            next_action = next(self.next_actions)
            next_action = find_best_line_to_reconnect(obs=observation, original_action=next_action)

        else:
            next_action = find_best_line_to_reconnect(obs=observation, original_action=self.action_space({}))
            idx_chosen = -1

        return next_action,idx_chosen

    def act( self, observation: grid2op.Observation.BaseObservation, reward: float, done: bool)\
            -> grid2op.Action.BaseAction:
        """Method of the agent to act.

        When the function selects a tuple action or triple action, the next steps are predetermined as
        well,i.e., all actions are returned sequentially.

        Args:
            observation: Grid2Op Observation
            reward: Reward of the previous action
            done: Whether the agent is done

        Returns: A suitable Grid2Op action

        """
        # We do not need done or the reward!

        action, _ = self.act_with_id(observation)
        return action

    def reset(self, obs: grid2op.Observation.BaseObservation):
        """ Resetting the agent.

        Args:
            obs:

        Returns:

        """
        self.next_actions = None

    def __collect_action(self, this_directory_path: str, action_space_path: Union[Path, List[Path]]) -> np.ndarray:
        """Check the action space path for the different action set.

        Args:
            this_directory_path: Directory of the submission files
            action_space_path: Optional action space path

        Returns:

        """
        actions = None
        if isinstance(action_space_path, Path):
            if action_space_path.is_file():
                logging.info(f"action_space_path {action_space_path} is a file and will be loaded.")
                actions = np.load(str(action_space_path))
            elif action_space_path.is_dir():
                logging.info(
                    f"action_space_path {action_space_path} is a path. All available action files "
                    f" will be loaded."
                )
                all_action_files = [
                    act for act in os.listdir(action_space_path) if "actions" in act and ".npy" in act
                ]

                if not all_action_files:
                    raise FileNotFoundError("No actions files were found!")

                loaded_files = []
                for act in all_action_files:
                    if "actions" in act and ".npy" in act:
                        loaded_files.append(np.load(action_space_path / act))

                actions = np.concatenate(loaded_files, axis=0)

        elif isinstance(action_space_path, list):
            logging.info(f"action_space_path {action_space_path} is a list containing multiple actions.")
            for act_path in action_space_path:
                if isinstance(act_path, Path):
                    assert act_path.is_file()
                else:
                    os.path.isfile(act_path)
            loaded_files = [np.load(str(act_path)) for act_path in action_space_path]
            actions = np.concatenate(loaded_files, axis=0)
        else:
            raise ValueError(
                f"The action_space_path variable {action_space_path} does neither consist of a single "
                f"action nor of a path where actions can be found."
            )

        return actions

    def __get_actions(self, obs: grid2op.Observation.BaseObservation):
        """ This method conducts depending on the underlying model the action method

        Args:
            obs: Input of the Grid2op Environment

        Returns: action

        """
        if self.run_with_tf:
            if isinstance(self.model, Model):
                # Newer Junior or Senior model
                sorted_actions = self.__get_keras_actions_model(obs=obs)

            else:
                # Older Model from Ray<2.4
                sorted_actions = self.__get_tf_actions(obs=obs)
        else:
            sorted_actions = self.__get_torch_actions_model(obs=obs)

        return sorted_actions

    def __get_keras_actions_model(self, obs: grid2op.Observation.BaseObservation):
        """Method to get the keras actions:

        Args:
            obs: Current observations

        Returns: numpy with the sorted actions:

        """
        # Select subset if wanted
        if isinstance(self.subset, list):
            model_input = obs.to_vect()[self.subset]
        elif self.subset:
            model_input = obs_to_vect(obs, False)
        else:
            model_input = obs.to_vect()

        if self.scaler:
            model_input = self.scaler.transform(model_input.reshape(1, -1))

        model_input = model_input.reshape((1, -1))

        if isinstance(self.model, tf.keras.models.Sequential):
            # Junior Model: Sequential model
            action_probability = self.model.predict(model_input, verbose=0)
        else:
            # Senior Model: tensorflow functional model:
            action_probability_pre_softmax, _ = self.model.predict(model_input, verbose=0)
            action_probability = tf.nn.softmax(action_probability_pre_softmax).numpy().reshape(-1)
        sorted_actions = action_probability.argsort()[::-1]
        return sorted_actions.reshape(-1)


    def __get_tf_actions(self, obs: grid2op.Observation.BaseObservation):
            """Method to get the tf actions:

            Args:
                obs: Current observations

            Returns: sorted numpy array with actions

            """
            # Select subset if wanted
            if isinstance(self.subset, list):
                model_input = obs.to_vect()[self.subset]
            elif self.subset:
                model_input = obs_to_vect(obs, False)
            else:
                model_input = obs.to_vect()

            if self.scaler:
                model_input = self.scaler.transform(model_input.reshape(1, -1)).reshape(
                    -1,
                )

            f = self.model.signatures["serving_default"]
            out = f(
                observations=tf.convert_to_tensor(model_input.reshape(1, -1)),
                timestep=tf.convert_to_tensor(0, dtype=tf.int64),
                is_training=tf.convert_to_tensor(False),
            )

            # Collect the softmax over all actions
            try:
                prob_of_action = (
                    tf.nn.softmax(out["action_dist_inputs"])
                    .numpy()
                    .reshape(
                        -1,
                    )
                )
            except AttributeError:
                poa = tf.nn.softmax(out["action_dist_inputs"])
                prob_of_action = poa.eval(session=tf.compat.v1.Session()).reshape(-1, )

            sorted_actions = prob_of_action.argsort()[::-1]
            return sorted_actions


    def __get_torch_actions_model(self, obs: grid2op.Observation.BaseObservation):
            """Method to get the torch actions (for PyTorch model).

            Args:
                obs: Current observations

            Returns: numpy array with the sorted actions

            """
            # Select subset if wanted
            if isinstance(self.subset, list):
                model_input = obs.to_vect()[self.subset]
            elif self.subset:
                model_input = obs_to_vect(obs, False)
            else:
                model_input = obs.to_vect()

            if self.scaler:
                model_input = self.scaler.transform(model_input.reshape(1, -1))

            model_input = model_input.reshape((1, -1))
            # Convert to PyTorch tensor
            model_input = torch.tensor(model_input, dtype=torch.float32)
            with torch.no_grad():
                output = self.model(model_input)
                action_probabilities = torch.softmax(output, dim=1).numpy().reshape(-1) #need to be checked

            sorted_actions = action_probabilities.argsort()[::-1]
            return sorted_actions

    def __generate_config(self, checkpoint):
        """
        Generate the configuration dictionary from a checkpoint dict.

        Args:
            checkpoint (dict): The checkpoint dictionary containing 'state_dict' and 'model_config'.

        Returns:
            dict: A dictionary containing the 'custom_config' with activation, initializer, and layer sizes.
        """
        # Initialize custom_config dictionary
        custom_config = {}

        # Extract 'model_config' from the checkpoint
        model_config = checkpoint.get('model_config', None)

        # Check if 'model_config' is not empty and extract values
        if model_config:
            activation = model_config.get('activation', 'relu')
            initializer = model_config.get('initializer', 'O')
            layer1 = model_config.get('layer1', 1000)
            layer2 = model_config.get('layer2', 1000)
            layer3 = model_config.get('layer3', 1000)
            layer4 = model_config.get('layer4', 1000)
        else:
            # Use default values if 'model_config' is empty
            activation = 'relu'
            initializer = 'O'
            layer1 = 1000
            layer2 = 1000
            layer3 = 1000
            layer4 = 1000

        custom_config['activation'] = activation
        custom_config['initializer'] = initializer
        custom_config['layer1'] = layer1
        custom_config['layer2'] = layer2
        custom_config['layer3'] = layer3
        custom_config['layer4'] = layer4

        # Extract the 'state_dict' from the checkpoint
        state_dict = checkpoint.get('state_dict', {})
        if not state_dict:
            raise ValueError("The checkpoint does not contain a 'state_dict'.")

        return {'custom_config': custom_config}



def make_agent(env, this_directory_path):
    my_agent = MyAgent(
        action_space=env.action_space,
        model_path=Path(this_directory_path) / "model",
        this_directory_path=Path(this_directory_path),
        action_space_path=Path(this_directory_path) / "actions",
        subset=True,
    )
    return my_agent



"""In this file, we do the following thing repeatedly:
    1. choose a scenario
    2. while not game over:
    3.     if not overflow:
    4.         step a "reconnect disconnected line" or "do nothing" action
    5.     else:
    6.         search a greedy action to minimize the max rho (~60k possible actions)
    7.         save the tuple of (None, observation, action) to a csv file.

Credit: The methods are the enhanced methods of the original code, see
@https://github.com/AsprinChina/L2RPN_NIPS_2020_a_PPO_Solution
"""
import logging
import os
import sys
import time
from multiprocessing import Pool
from pathlib import Path
from typing import Union

import defopt
import grid2op
import numpy as np
from grid2op.Backend import PandaPowerBackend
from grid2op.Environment import BaseEnv

from curriculumagent.teacher.submodule.common import save_sample_new
from curriculumagent.common.utilities import find_best_line_to_reconnect
from curriculumagent.teacher.submodule.encoded_action import EncodedTopologyAction
from curriculumagent.teacher.submodule.topology_action_search import topology_search_topk


def collect_general_experience(
        save_path: Path,
        env_name_path: Union[str, Path],
        chronics_path: str,
        seed: int,
        rho_acting_threshold: float = 0.925,
        top_k: int = 125,
        chronic_limit: int = None,
):
    """Collect general teacher_experience by running a greedy search whenever the agent encounters a high power load.

    Args:
        save_path: The path to save the teacher_experience to.
        env_name_path: The name or path of the environment used.
        chronics_path: The path to the chronics of the environment.
        seed: The seed of the environment.
        rho_acting_threshold: When the environments max_rho goes above this threshold the agent begins to act.
        top_k: How many of the top best actions should be saved in the teacher_experience file at save_path.
        chronic_limit: How many chronics to run. Runs endless if set to None.

    Returns:
        None.
    """
    if isinstance(env_name_path, Path):
        env_name_path = str(env_name_path)

    # Setup environment with best possible backend
    try:
        # if lightsim2grid is available, use it.
        from lightsim2grid import LightSimBackend

        backend = LightSimBackend()
    except ImportError:  # noqa
        backend = PandaPowerBackend()
        logging.warning("Not using lightsim2grid! Operation will be slow!")
    if chronics_path:
        # Use env_name_path as path together with chronics_path
        env: BaseEnv = grid2op.make(dataset=env_name_path, chronics_path=chronics_path, backend=backend)
    else:
        # Use env_name_path as name
        env: BaseEnv = grid2op.make(dataset=env_name_path, backend=backend)

    all_actions = env.action_space.get_all_unitary_topologies_set(env.action_space)
    np.random.seed(seed)
    env.chronics_handler.shuffle(shuffler=lambda x: x[np.random.choice(len(x), size=len(x), replace=False)])
    num_chronics = len(os.listdir(chronics_path))
    if chronic_limit:
        num_chronics = min(num_chronics, chronic_limit)
    for _ in range(num_chronics):
        env.reset()  # env.reset() loads the next chronic
        dst_step = 0
        logging.info(f"Scenario to test is [{env.chronics_handler.get_name()}]，start from step-{dst_step:d}... ...")
        env.fast_forward_chronics(dst_step)
        obs, done = env.get_obs(), False
        while not done:
            if obs.rho.max() >= rho_acting_threshold:
                best_actions = topology_search_topk(env, obs, all_actions, top_k=125)
                if len(best_actions) > 0:
                    action = best_actions[0][1]
                    best_actions = [(ri, EncodedTopologyAction(act)) for ri, act in best_actions]
                    obs_, _, done, info = env.step(action)
                    save_sample_new(Path(save_path), best_actions, obs, obs_, top_k=top_k)
                    obs = obs_
                else:
                    action = env.action_space({})
                    action = find_best_line_to_reconnect(obs, action)
                    obs, _, done, info = env.step(action)
            else:
                action = env.action_space({})
                action = find_best_line_to_reconnect(obs, action)
                obs, _, done, info = env.step(action)
            if done and obs.current_step < obs.max_step:
                logging.info(
                    f"Game over at t={obs.current_step}/{obs.max_step} at chronic: {env.chronics_handler.get_name()}"
                )


def run_general_teacher(
        save_path: Path,
        env_name_path: Union[Path, str] = "l2rpn_neurips_2020_track1_small",
        rho_acting_threshold: float = 0.925,
        *,
        n_episodes: int = 100,
        limit_chronics: int = None,
        top_k: int = 125,
        jobs: int = os.cpu_count(),
        seed: int = 42,
):
    """This teacher is trying to collect useful actions by doing topology search while running an
    environment when an overflow occurs.

    Args:
        save_path: Path to save the teacher_experience file to.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        rho_acting_threshold: When the environments max_rho goes above this threshold the agent begins to act.
        n_episodes: Number of episodes (run through all scenarios) to run.
        limit_chronics: Limit the number of chronics to run through.
        top_k: How many of the top best actions should be saved in the teacher_experience file at save_path.
        jobs: Number of jobs to use for parallel teacher_experience collection.
        seed: The random seed to use.

    Returns:
        None.
    """

    # Make sure we can initialize the environment
    # This also makes sure that the environment actually exits or gets downloaded
    env: BaseEnv = grid2op.make(env_name_path)
    chronics_path = env.chronics_handler.path
    if chronics_path is None:
        raise ValueError(f"Can't determine chronics path of given environment {env_name_path}")

    # Parallel implementation
    start = time.time()
    tasks = []
    for episode in range(n_episodes):
        tasks.append(
            (save_path, env_name_path, chronics_path, seed + episode, rho_acting_threshold, top_k, limit_chronics)
        )
    if jobs == 1:
        # This makes debugging easier since we don't fork into multiple processes
        logging.info(f"The following {len(tasks)} tasks will executed sequentially: {tasks}")
        for task in tasks:
            collect_general_experience(*task)
    else:
        logging.info(f"The following {len(tasks)} tasks will be distributed to a pool of {jobs} workers: {tasks}")
        with Pool(jobs) as p:
            p.starmap(collect_general_experience, tasks)
    end = time.time()
    elapsed = end - start
    logging.info(f"Parallel time: {elapsed}s")


